import argparse, os
import numba as nb

# Parse command-line arguments
#   TODO: Will be inside run() once Python/Numba adapter is integrated
parser = argparse.ArgumentParser(description="MC/DC: Monte Carlo Dynamic Code")
parser.add_argument(
    "--mode", type=str, help="Run mode", choices=["python", "numba"], default="python"
)
parser.add_argument("--N_particle", type=int, help="Number of particles")
parser.add_argument("--output", type=str, help="Output file name")
parser.add_argument("--progress_bar", default=False, action="store_true")
parser.add_argument("--no-progress_bar", dest="progress_bar", action="store_false")
args, unargs = parser.parse_known_args()

# Set mode
#   TODO: Will be inside run() once Python/Numba adapter is integrated
mode = args.mode
if mode == "python":
    nb.config.DISABLE_JIT = True
elif mode == "numba":
    nb.config.DISABLE_JIT = False

import h5py
import numpy as np

from mpi4py import MPI
from scipy.stats import qmc

import mcdc.kernel as kernel
import mcdc.type_ as type_

from mcdc.constant import *
from mcdc.loop import loop_fixed_source, loop_eigenvalue, loop_iqmc
from mcdc.print_ import print_banner, print_msg, print_runtime, print_header_eigenvalue

# Get input_deck
import mcdc.global_ as mcdc_

input_deck = mcdc_.input_deck


def run():
    # Override input deck with command-line argument, if given
    if args.N_particle is not None:
        input_deck.setting["N_particle"] = args.N_particle
    if args.output is not None:
        input_deck.setting["output_name"] = args.output
    if args.progress_bar is not None:
        input_deck.setting["progress_bar"] = args.progress_bar

    # Start timer
    total_start = MPI.Wtime()

    # Preparation
    #   Set up and get the global variable container `mcdc` based on
    #   input deck
    preparation_start = MPI.Wtime()
    mcdc = prepare()
    mcdc["runtime_preparation"] = MPI.Wtime() - preparation_start

    # Print banner, hardware configuration, and header
    print_banner(mcdc)
    print_msg(" Now running TNT...")
    if mcdc["setting"]["mode_eigenvalue"]:
        print_header_eigenvalue(mcdc)

    # Run simulation
    simulation_start = MPI.Wtime()
    if mcdc["technique"]["iQMC"]:
        loop_iqmc(mcdc)
    elif mcdc["setting"]["mode_eigenvalue"]:
        loop_eigenvalue(mcdc)
    else:
        loop_fixed_source(mcdc)
    mcdc["runtime_simulation"] = MPI.Wtime() - simulation_start

    # Output: generate hdf5 output files
    output_start = MPI.Wtime()
    generate_hdf5(mcdc)
    mcdc["runtime_output"] = MPI.Wtime() - output_start

    # Stop timer
    MPI.COMM_WORLD.Barrier()
    mcdc["runtime_total"] = MPI.Wtime() - total_start

    # Closout
    closeout(mcdc)


def prepare():
    """
    Preparing the MC transport simulation:
      (1) Adapt kernels
      (2) Make types
      (3) Create and set up global variable container `mcdc`
    """

    # =========================================================================
    # Adapt kernels
    # =========================================================================

    kernel.adapt_rng(nb.config.DISABLE_JIT)

    # =========================================================================
    # Make types
    # =========================================================================

    type_.make_type_particle(input_deck)
    type_.make_type_particle_record(input_deck)
    type_.make_type_nuclide(input_deck)
    type_.make_type_material(input_deck)
    type_.make_type_surface(input_deck)
    type_.make_type_cell(input_deck)
    type_.make_type_universe(input_deck)
    type_.make_type_lattice(input_deck)
    type_.make_type_source(input_deck)
    type_.make_type_tally(input_deck)
    type_.make_type_setting(input_deck)
    type_.make_type_uq_tally(input_deck)
    type_.make_type_uq(input_deck)
    type_.make_type_technique(input_deck)
    type_.make_type_global(input_deck)

    # =========================================================================
    # Create the global variable container
    #   TODO: Better alternative?
    # =========================================================================

    mcdc = np.zeros(1, dtype=type_.global_)[0]

    # Now, set up the global variable container

    # Get modes
    mode_CE = input_deck.setting["mode_CE"]
    mode_MG = input_deck.setting["mode_MG"]

    # =========================================================================
    # Nuclides
    # =========================================================================

    N_nuclide = len(input_deck.nuclides)
    for i in range(N_nuclide):
        # General data
        for name in ["ID", "fissionable", "sensitivity", "sensitivity_ID", "dsm_Np"]:
            mcdc["nuclides"][i][name] = input_deck.nuclides[i][name]

        # MG data
        if mode_MG:
            for name in [
                "G",
                "J",
                "speed",
                "decay",
                "total",
                "capture",
                "scatter",
                "fission",
                "nu_s",
                "nu_f",
                "nu_p",
                "nu_d",
                "chi_s",
                "chi_p",
                "chi_d",
            ]:
                mcdc["nuclides"][i][name] = input_deck.nuclides[i][name]

        # CE data (load data from XS library)
        dir_name = os.getenv("MCDC_XSLIB")
        if mode_CE:
            nuc_name = input_deck.nuclides[i]["name"]
            with h5py.File(dir_name + "/" + nuc_name + ".h5", "r") as f:
                # Atomic weight ratio
                mcdc["nuclides"][i]["A"] = f["A"][()]
                # Energy grids
                for name in [
                    "E_xs",
                    "E_nu_p",
                    "E_nu_d",
                    "E_chi_p",
                    "E_chi_d1",
                    "E_chi_d2",
                    "E_chi_d3",
                    "E_chi_d4",
                    "E_chi_d5",
                    "E_chi_d6",
                ]:
                    mcdc["nuclides"][i]["N" + name] = len(f[name][:])
                    mcdc["nuclides"][i][name][: len(f[name][:])] = f[name][:]

                # XS
                for name in ["capture", "scatter", "fission"]:
                    mcdc["nuclides"][i]["ce_" + name][: len(f[name][:])] = f[name][:]
                    mcdc["nuclides"][i]["ce_total"][: len(f[name][:])] += f[name][:]

                # Fission production
                mcdc["nuclides"][i]["ce_nu_p"][: len(f["nu_p"][:])] = f["nu_p"][:]
                for j in range(6):
                    mcdc["nuclides"][i]["ce_nu_d"][j][: len(f["nu_d"][j, :])] = f[
                        "nu_d"
                    ][j, :]

                # Fission spectrum
                mcdc["nuclides"][i]["ce_chi_p"][: len(f["chi_p"][:])] = f["chi_p"][:]
                for j in range(6):
                    mcdc["nuclides"][i]["ce_chi_d%i" % (j + 1)][
                        : len(f["chi_d%i" % (j + 1)][:])
                    ] = f["chi_d%i" % (j + 1)][:]

                # Decay
                mcdc["nuclides"][i]["ce_decay"][: len(f["decay_rate"][:])] = f[
                    "decay_rate"
                ][:]

    # =========================================================================
    # Materials
    # =========================================================================

    N_material = len(input_deck.materials)
    for i in range(N_material):
        for name in type_.material.names:
            if name in ["nuclide_IDs", "nuclide_densities"]:
                mcdc["materials"][i][name][
                    : mcdc["materials"][i]["N_nuclide"]
                ] = input_deck.materials[i][name]
            else:
                mcdc["materials"][i][name] = input_deck.materials[i][name]

    # =========================================================================
    # Surfaces
    # =========================================================================

    N_surface = len(input_deck.surfaces)
    for i in range(N_surface):
        for name in type_.surface.names:
            if name not in ["J", "t"]:
                mcdc["surfaces"][i][name] = input_deck.surfaces[i][name]

        # Variables with possible different sizes
        for name in ["J", "t"]:
            N = len(input_deck.surfaces[i][name])
            mcdc["surfaces"][i][name][:N] = input_deck.surfaces[i][name]

    # =========================================================================
    # Cells
    # =========================================================================

    N_cell = len(input_deck.cells)
    for i in range(N_cell):
        for name in type_.cell.names:
            if name not in ["surface_IDs", "positive_flags"]:
                mcdc["cells"][i][name] = input_deck.cells[i][name]

        # Variables with possible different sizes
        for name in ["surface_IDs", "positive_flags"]:
            N = len(input_deck.cells[i][name])
            mcdc["cells"][i][name][:N] = input_deck.cells[i][name]

    # =========================================================================
    # Universes
    # =========================================================================

    N_universe = len(input_deck.universes)
    for i in range(N_universe):
        for name in type_.universe.names:
            if name not in ["cell_IDs"]:
                mcdc["universes"][i][name] = input_deck.universes[i][name]

        # Variables with possible different sizes
        for name in ["cell_IDs"]:
            N = mcdc["universes"][i]["N_cell"]
            mcdc["universes"][i][name][:N] = input_deck.universes[i][name]

    # =========================================================================
    # Lattices
    # =========================================================================

    N_lattice = len(input_deck.lattices)
    for i in range(N_lattice):
        # Mesh
        for name in type_.mesh_uniform.names:
            mcdc["lattices"][i]["mesh"][name] = input_deck.lattices[i]["mesh"][name]

        # Universe IDs
        Nx = mcdc["lattices"][i]["mesh"]["Nx"]
        Ny = mcdc["lattices"][i]["mesh"]["Ny"]
        Nz = mcdc["lattices"][i]["mesh"]["Nz"]
        mcdc["lattices"][i]["universe_IDs"][:Nx, :Ny, :Nz] = input_deck.lattices[i][
            "universe_IDs"
        ]

    # =========================================================================
    # Source
    # =========================================================================

    N_source = len(input_deck.sources)
    for i in range(N_source):
        for name in type_.source.names:
            mcdc["sources"][i][name] = input_deck.sources[i][name]

    # Normalize source probabilities
    tot = 0.0
    for S in mcdc["sources"]:
        tot += S["prob"]
    for S in mcdc["sources"]:
        S["prob"] /= tot

    # =========================================================================
    # Tally
    # =========================================================================

    for name in type_.tally.names:
        if name not in ["score", "mesh"]:
            mcdc["tally"][name] = input_deck.tally[name]
    # Set mesh
    for name in type_.mesh_names:
        mcdc["tally"]["mesh"][name] = input_deck.tally["mesh"][name]

    # =========================================================================
    # Setting
    # =========================================================================

    for name in type_.setting.names:
        mcdc["setting"][name] = input_deck.setting[name]

    # Check if time boundary is above the final tally mesh time grid
    if mcdc["setting"]["time_boundary"] > mcdc["tally"]["mesh"]["t"][-1]:
        mcdc["setting"]["time_boundary"] = mcdc["tally"]["mesh"]["t"][-1]

    if input_deck.technique["iQMC"]:
        if len(mcdc["technique"]["iqmc"]["mesh"]["t"]) - 1 > 1:
            if (
                mcdc["setting"]["time_boundary"]
                > input_deck.technique["iqmc"]["mesh"]["t"][-1]
            ):
                mcdc["setting"]["time_boundary"] = input_deck.technique["iqmc"]["mesh"][
                    "t"
                ][-1]
    # =========================================================================
    # Technique
    # =========================================================================

    # Flags
    for name in [
        "weighted_emission",
        "implicit_capture",
        "population_control",
        "weight_window",
        "weight_roulette",
        "iQMC",
        "IC_generator",
        "branchless_collision",
        "uq",
    ]:
        mcdc["technique"][name] = input_deck.technique[name]

    # =========================================================================
    # Population control
    # =========================================================================

    # Population control technique (PCT)
    mcdc["technique"]["pct"] = input_deck.technique["pct"]
    mcdc["technique"]["pc_factor"] = input_deck.technique["pc_factor"]

    # =========================================================================
    # IC generator
    # =========================================================================

    for name in [
        "IC_N_neutron",
        "IC_N_precursor",
        "IC_neutron_density",
        "IC_neutron_density_max",
        "IC_precursor_density",
        "IC_precursor_density_max",
    ]:
        mcdc["technique"][name] = input_deck.technique[name]

    # =========================================================================
    # Weight window (WW)
    # =========================================================================

    # WW mesh
    for name in type_.mesh_names[:-1]:
        mcdc["technique"]["ww_mesh"][name] = input_deck.technique["ww_mesh"][name]

    # WW windows
    mcdc["technique"]["ww"] = input_deck.technique["ww"]
    mcdc["technique"]["ww_width"] = input_deck.technique["ww_width"]

    # =========================================================================
    # Weight roulette
    # =========================================================================

    # Threshold
    mcdc["technique"]["wr_threshold"] = input_deck.technique["wr_threshold"]

    # Survival probability
    mcdc["technique"]["wr_chance"] = input_deck.technique["wr_chance"]

    # =========================================================================
    # Quasi Monte Carlo
    # =========================================================================

    for name in type_.technique["iqmc"].names:
        if name not in [
            "mesh",
            "res",
            "lds",
            "generator",
            "sweep_counter",
            "total_source",
            "w_min",
            "score_list",
            "score",
        ]:
            mcdc["technique"]["iqmc"][name] = input_deck.technique["iqmc"][name]

    if input_deck.technique["iQMC"]:
        # pass in mesh
        iqmc = mcdc["technique"]["iqmc"]
        iqmc["mesh"]["x"] = input_deck.technique["iqmc"]["mesh"]["x"]
        iqmc["mesh"]["y"] = input_deck.technique["iqmc"]["mesh"]["y"]
        iqmc["mesh"]["z"] = input_deck.technique["iqmc"]["mesh"]["z"]
        iqmc["mesh"]["t"] = input_deck.technique["iqmc"]["mesh"]["t"]
        # pass in score list
        for name, value in input_deck.technique["iqmc"]["score_list"].items():
            iqmc["score_list"][name] = value
        # pass in initial tallies
        for name, value in input_deck.technique["iqmc"]["score"].items():
            mcdc["technique"]["iqmc"]["score"][name] = value
        # LDS generator
        iqmc["generator"] = input_deck.technique["iqmc"]["generator"]
        # minimum particle weight
        iqmc["w_min"] = 1e-13
        # variables to generate samples
        scramble = iqmc["scramble"]
        N_dim = iqmc["N_dim"]
        seed = iqmc["seed"]
        N_particle = mcdc["setting"]["N_particle"]

        mcdc["mpi_size"] = MPI.COMM_WORLD.Get_size()
        mcdc["mpi_rank"] = MPI.COMM_WORLD.Get_rank()
        kernel.distribute_work(N_particle, mcdc)
        N_work = int(mcdc["mpi_work_size"])
        N_start = int(mcdc["mpi_work_start"])

        # generate LDS
        if input_deck.technique["iqmc"]["generator"] == "sobol":
            sampler = qmc.Sobol(d=N_dim, scramble=scramble)
            # skip the first entry in Sobol sequence because its 0.0
            # skip the second because it maps to ux = 0.0
            sampler.fast_forward(2)
            sampler.fast_forward(N_start)
            iqmc["lds"] = sampler.random(N_work)
        if input_deck.technique["iqmc"]["generator"] == "halton":
            sampler = qmc.Halton(d=N_dim, scramble=scramble, seed=seed)
            # skip the first entry in Halton sequence because its 0.0
            sampler.fast_forward(1)
            sampler.fast_forward(N_start)
            iqmc["lds"] = sampler.random(N_work)
        if input_deck.technique["iqmc"]["generator"] == "random":
            np.random.seed(seed)
            seeds = np.random.randint(1e6, size=mcdc["mpi_size"])
            np.random.seed(seeds[mcdc["mpi_rank"]])
            iqmc["lds"] = np.random.random((N_work, N_dim))

    # =========================================================================
    # Derivative Source Method
    # =========================================================================

    # Threshold
    mcdc["technique"]["dsm_order"] = input_deck.technique["dsm_order"]

    # =========================================================================
    # Variance Deconvolution - UQ
    # =========================================================================
    if mcdc["technique"]["uq"]:
        # Assumes that all tallies will also be uq tallies
        for name in type_.uq_tally.names:
            if name != "score":
                mcdc["technique"]["uq_tally"][name] = input_deck.tally[name]

        M = len(input_deck.uq_deltas["materials"])
        for i in range(M):
            idm = input_deck.uq_deltas["materials"][i]["ID"]
            mcdc["technique"]["uq_"]["materials"][i]["info"]["ID"] = idm
            mcdc["technique"]["uq_"]["materials"][i]["info"][
                "distribution"
            ] = input_deck.uq_deltas["materials"][i]["distribution"]
            for name in input_deck.uq_deltas["materials"][i]["flags"]:
                mcdc["technique"]["uq_"]["materials"][i]["flags"][name] = True
                mcdc["technique"]["uq_"]["materials"][i]["delta"][
                    name
                ] = input_deck.uq_deltas["materials"][i][name]
            flags = mcdc["technique"]["uq_"]["materials"][i]["flags"]
            if flags["capture"] or flags["scatter"] or flags["fission"]:
                flags["total"] = True
                flags["speed"] = True
            if flags["nu_p"] or flags["nu_d"]:
                flags["nu_f"] = True
            if mcdc["materials"][idm]["N_nuclide"] > 1:
                for name in type_.uq_mat.names:
                    mcdc["technique"]["uq_"]["materials"][i]["mean"][
                        name
                    ] = input_deck.materials[idm][name]

        N = len(input_deck.uq_deltas["nuclides"])
        for i in range(N):
            mcdc["technique"]["uq_"]["nuclides"][i]["info"][
                "distribution"
            ] = input_deck.uq_deltas["nuclides"][i]["distribution"]
            idn = input_deck.uq_deltas["nuclides"][i]["ID"]
            mcdc["technique"]["uq_"]["nuclides"][i]["info"]["ID"] = idn
            for name in type_.uq_nuc.names:
                mcdc["technique"]["uq_"]["nuclides"][i]["mean"][
                    name
                ] = input_deck.nuclides[idn][name]
            for name in input_deck.uq_deltas["nuclides"][i]["flags"]:
                mcdc["technique"]["uq_"]["nuclides"][i]["flags"][name] = True
                mcdc["technique"]["uq_"]["nuclides"][i]["delta"][
                    name
                ] = input_deck.uq_deltas["nuclides"][i][name]
            flags = mcdc["technique"]["uq_"]["nuclides"][i]["flags"]
            if flags["capture"] or flags["scatter"] or flags["fission"]:
                flags["total"] = True
            if flags["nu_p"] or flags["nu_d"]:
                flags["nu_f"] = True

    # =========================================================================
    # MPI
    # =========================================================================

    # MPI parameters
    mcdc["mpi_size"] = MPI.COMM_WORLD.Get_size()
    mcdc["mpi_rank"] = MPI.COMM_WORLD.Get_rank()
    mcdc["mpi_master"] = mcdc["mpi_rank"] == 0

    # Distribute work to MPI ranks
    kernel.distribute_work(mcdc["setting"]["N_particle"], mcdc)

    # =========================================================================
    # Particle banks
    # =========================================================================

    # Particle bank tags
    mcdc["bank_active"]["tag"] = "active"
    mcdc["bank_census"]["tag"] = "census"
    mcdc["bank_source"]["tag"] = "source"

    # IC generator banks
    if mcdc["technique"]["IC_generator"]:
        mcdc["technique"]["IC_bank_neutron_local"]["tag"] = "neutron"
        mcdc["technique"]["IC_bank_precursor_local"]["tag"] = "precursor"
        mcdc["technique"]["IC_bank_neutron"]["tag"] = "neutron"
        mcdc["technique"]["IC_bank_precursor"]["tag"] = "precursor"

    # =========================================================================
    # Eigenvalue (or fixed-source)
    # =========================================================================

    # Initial guess
    mcdc["k_eff"] = mcdc["setting"]["k_init"]

    # Activate tally scoring for fixed-source
    if not mcdc["setting"]["mode_eigenvalue"]:
        mcdc["cycle_active"] = True

    # All active eigenvalue cycle?
    elif mcdc["setting"]["N_inactive"] == 0:
        mcdc["cycle_active"] = True

    # =========================================================================
    # Source file
    # =========================================================================

    if mcdc["setting"]["source_file"]:
        with h5py.File(mcdc["setting"]["source_file_name"], "r") as f:
            # Get source particle size
            N_particle = f["particles_size"][()]

            # Redistribute work
            kernel.distribute_work(N_particle, mcdc)
            N_local = mcdc["mpi_work_size"]
            start = mcdc["mpi_work_start"]
            end = start + N_local

            # Add particles to source bank
            mcdc["bank_source"]["particles"][:N_local] = f["particles"][start:end]
            mcdc["bank_source"]["size"] = N_local

    # =========================================================================
    # IC file
    # =========================================================================

    if mcdc["setting"]["IC_file"]:
        with h5py.File(mcdc["setting"]["IC_file_name"], "r") as f:
            # =================================================================
            # Set neutron source
            # =================================================================

            # Get source particle size
            N_particle = f["IC/neutrons_size"][()]

            # Redistribute work
            kernel.distribute_work(N_particle, mcdc)
            N_local = mcdc["mpi_work_size"]
            start = mcdc["mpi_work_start"]
            end = start + N_local

            # Add particles to source bank
            mcdc["bank_source"]["particles"][:N_local] = f["IC/neutrons"][start:end]
            mcdc["bank_source"]["size"] = N_local

            # =================================================================
            # Set precursor source
            # =================================================================

            # Get source particle size
            N_precursor = f["IC/precursors_size"][()]

            # Redistribute work
            kernel.distribute_work(N_precursor, mcdc, True)  # precursor = True
            N_local = mcdc["mpi_work_size_precursor"]
            start = mcdc["mpi_work_start_precursor"]
            end = start + N_local

            # Add particles to source bank
            mcdc["bank_precursor"]["precursors"][:N_local] = f["IC/precursors"][
                start:end
            ]
            mcdc["bank_precursor"]["size"] = N_local

            # Set precursor strength
            if N_precursor > 0 and N_particle > 0:
                mcdc["precursor_strength"] = mcdc["bank_precursor"]["precursors"][0][
                    "w"
                ]

    return mcdc


def dictlist_to_h5group(dictlist, input_group, name):
    main_group = input_group.create_group(name + "s")
    for item in dictlist:
        group = main_group.create_group(name + "_%i" % item["ID"])
        dict_to_h5group(item, group)


def dict_to_h5group(dict_, group):
    for k, v in dict_.items():
        if type(v) == dict:
            dict_to_h5group(dict_[k], group.create_group(k))
        elif v is None:
            next
        else:
            group[k] = v


def generate_hdf5(mcdc):
    if mcdc["mpi_master"]:
        if mcdc["setting"]["progress_bar"]:
            print_msg("")
        print_msg(" Generating output HDF5 files...")

        with h5py.File(mcdc["setting"]["output_name"] + ".h5", "w") as f:
            # Input deck
            if mcdc["setting"]["save_input_deck"]:
                input_group = f.create_group("input_deck")
                dictlist_to_h5group(input_deck.nuclides, input_group, "nuclide")
                dictlist_to_h5group(input_deck.materials, input_group, "material")
                dictlist_to_h5group(input_deck.surfaces, input_group, "surface")
                dictlist_to_h5group(input_deck.cells, input_group, "cell")
                dictlist_to_h5group(input_deck.universes, input_group, "universe")
                dictlist_to_h5group(input_deck.lattices, input_group, "lattice")
                dictlist_to_h5group(input_deck.sources, input_group, "source")
                dict_to_h5group(input_deck.tally, input_group.create_group("tally"))
                dict_to_h5group(input_deck.setting, input_group.create_group("setting"))
                dict_to_h5group(
                    input_deck.technique, input_group.create_group("technique")
                )

            # Tally
            T = mcdc["tally"]
            f.create_dataset("tally/grid/t", data=T["mesh"]["t"])
            f.create_dataset("tally/grid/x", data=T["mesh"]["x"])
            f.create_dataset("tally/grid/y", data=T["mesh"]["y"])
            f.create_dataset("tally/grid/z", data=T["mesh"]["z"])
            f.create_dataset("tally/grid/mu", data=T["mesh"]["mu"])
            f.create_dataset("tally/grid/azi", data=T["mesh"]["azi"])
            f.create_dataset("tally/grid/g", data=T["mesh"]["g"])

            # Scores
            for name in T["score"].dtype.names:
                if mcdc["tally"][name]:
                    name_h5 = name.replace("_", "-")
                    f.create_dataset(
                        "tally/" + name_h5 + "/mean",
                        data=np.squeeze(T["score"][name]["mean"]),
                    )
                    f.create_dataset(
                        "tally/" + name_h5 + "/sdev",
                        data=np.squeeze(T["score"][name]["sdev"]),
                    )
                    if mcdc["technique"]["uq_tally"][name]:
                        mc_var = mcdc["technique"]["uq_tally"]["score"][name][
                            "batch_var"
                        ]
                        tot_var = mcdc["technique"]["uq_tally"]["score"][name][
                            "batch_bin"
                        ]
                        f.create_dataset(
                            "tally/" + name_h5 + "/uq_var",
                            data=np.squeeze(tot_var - mc_var),
                        )

            # Eigenvalues
            if mcdc["setting"]["mode_eigenvalue"]:
                if mcdc["technique"]["iQMC"]:
                    f.create_dataset("k_eff", data=mcdc["k_eff"])
                else:
                    N_cycle = mcdc["setting"]["N_cycle"]
                    f.create_dataset("k_cycle", data=mcdc["k_cycle"][:N_cycle])
                    f.create_dataset("k_mean", data=mcdc["k_avg_running"])
                    f.create_dataset("k_sdev", data=mcdc["k_sdv_running"])
                    f.create_dataset("global_tally/neutron/mean", data=mcdc["n_avg"])
                    f.create_dataset("global_tally/neutron/sdev", data=mcdc["n_sdv"])
                    f.create_dataset("global_tally/neutron/max", data=mcdc["n_max"])
                    f.create_dataset("global_tally/precursor/mean", data=mcdc["C_avg"])
                    f.create_dataset("global_tally/precursor/sdev", data=mcdc["C_sdv"])
                    f.create_dataset("global_tally/precursor/max", data=mcdc["C_max"])
                    if mcdc["setting"]["gyration_radius"]:
                        f.create_dataset(
                            "gyration_radius", data=mcdc["gyration_radius"][:N_cycle]
                        )

            # iQMC
            if mcdc["technique"]["iQMC"]:
                # dump iQMC mesh
                T = mcdc["technique"]
                f.create_dataset("iqmc/grid/t", data=T["iqmc"]["mesh"]["t"])
                f.create_dataset("iqmc/grid/x", data=T["iqmc"]["mesh"]["x"])
                f.create_dataset("iqmc/grid/y", data=T["iqmc"]["mesh"]["y"])
                f.create_dataset("iqmc/grid/z", data=T["iqmc"]["mesh"]["z"])
                # dump x,y,z scalar flux across all groups
                f.create_dataset(
                    "iqmc/tally/flux", data=np.squeeze(T["iqmc"]["score"]["flux"])
                )
                f.create_dataset(
                    "iqmc/tally/fission_source",
                    data=T["iqmc"]["score"]["fission-source"],
                )
                f.create_dataset(
                    "iqmc/tally/fission_power", data=T["iqmc"]["score"]["fission-power"]
                )
                f.create_dataset("iqmc/tally/source_constant", data=T["iqmc"]["source"])
                f.create_dataset(
                    "iqmc/tally/source_x", data=T["iqmc"]["score"]["tilt-x"]
                )
                f.create_dataset(
                    "iqmc/tally/source_y", data=T["iqmc"]["score"]["tilt-y"]
                )
                f.create_dataset(
                    "iqmc/tally/source_z", data=T["iqmc"]["score"]["tilt-z"]
                )
                f.create_dataset(
                    "iqmc/tally/source_xy", data=T["iqmc"]["score"]["tilt-xy"]
                )
                f.create_dataset(
                    "iqmc/tally/source_xz", data=T["iqmc"]["score"]["tilt-xz"]
                )
                f.create_dataset(
                    "iqmc/tally/source_yz", data=T["iqmc"]["score"]["tilt-yz"]
                )
                # iteration data
                f.create_dataset("iqmc/itteration_count", data=T["iqmc"]["itt"])
                f.create_dataset("iqmc/final_residual", data=T["iqmc"]["res"])
                f.create_dataset("iqmc/sweep_count", data=T["iqmc"]["sweep_counter"])
                if mcdc["setting"]["mode_eigenvalue"]:
                    f.create_dataset(
                        "iqmc/outter_itteration_count", data=T["iqmc"]["itt_outter"]
                    )
                    f.create_dataset(
                        "iqmc/outter_final_residual", data=T["iqmc"]["res_outter"]
                    )

            # Particle tracker
            if mcdc["setting"]["track_particle"]:
                with h5py.File(mcdc["setting"]["output"] + "_ptrack.h5", "w") as f:
                    N_track = mcdc["particle_track_N"]
                    f.create_dataset("tracks", data=mcdc["particle_track"][:N_track])

            # IC generator
            if mcdc["technique"]["IC_generator"]:
                Nn = mcdc["technique"]["IC_bank_neutron"]["size"]
                Np = mcdc["technique"]["IC_bank_precursor"]["size"]
                f.create_dataset(
                    "IC/neutrons",
                    data=mcdc["technique"]["IC_bank_neutron"]["particles"][:Nn],
                )
                f.create_dataset(
                    "IC/precursors",
                    data=mcdc["technique"]["IC_bank_precursor"]["precursors"][:Np],
                )
                f.create_dataset("IC/neutrons_size", data=Nn)
                f.create_dataset("IC/precursors_size", data=Np)
                f.create_dataset(
                    "IC/fission", data=mcdc["technique"]["IC_fission"] / Nn
                )

    # Save particle?
    if mcdc["setting"]["save_particle"]:
        # Gather source bank
        # TODO: Parallel HDF5 and mitigation of large data passing
        N = mcdc["bank_source"]["size"]
        neutrons = MPI.COMM_WORLD.gather(mcdc["bank_source"]["particles"][:N])

        # Master saves the particle
        if mcdc["mpi_master"]:
            # Remove unwanted particle fields
            neutrons = np.concatenate(neutrons[:])

            # Create dataset
            with h5py.File(mcdc["setting"]["output_name"] + ".h5", "a") as f:
                f.create_dataset("particles", data=neutrons[:])
                f.create_dataset("particles_size", data=len(neutrons[:]))


def closeout(mcdc):
    # Runtime
    if mcdc["mpi_master"]:
        with h5py.File(mcdc["setting"]["output_name"] + ".h5", "a") as f:
            for name in [
                "total",
                "preparation",
                "simulation",
                "output",
                "bank_management",
            ]:
                f.create_dataset(
                    "runtime/" + name, data=np.array([mcdc["runtime_" + name]])
                )

    print_runtime(mcdc)
    input_deck.reset()



import pytest


# Moves working directory to test file directory
@pytest.fixture(autouse=True)
def change_test_dir(request, monkeypatch):
    monkeypatch.chdir(request.fspath.dirname)


# Set mode (python, numba), mpi (srun, mpiexec), and number of ranks
def pytest_addoption(parser):
    parser.addoption(
        "--mode", action="store", default="python", choices=("python", "numba")
    )
    parser.addoption(
        "--mpi", action="store", default="srun", choices=("srun", "mpiexec")
    )
    parser.addoption("--N_rank", action="store", default="1")


@pytest.fixture(scope="session")
def mode(pytestconfig):
    return pytestconfig.getoption("mode")


@pytest.fixture(scope="session")
def mpi(pytestconfig):
    return pytestconfig.getoption("mpi")


@pytest.fixture(scope="session")
def N_rank(pytestconfig):
    return pytestconfig.getoption("N_rank")



try:
    # launches visualization window
    # must be inside this loop so it doesn't launch when the visualizer is imported
    from netgen.meshing import *
    from netgen.csg import *
    from ngsolve import Draw, Redraw  # just for visualization
    import distinctipy  # creates unlimited visually distinct colors for visualization

except ImportError as e:
    print("")
    print(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")
    print("MC/DC visualization error:")
    print("     Dependencies for visualization not installed")
    print("     To install optional dependencies needed for visualization:")
    print("         <pip install mcdc[viz]> ")
    print("")
    print(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")
    print("")


import tkinter as tk  # Tkinter is used to create the window for the time slider and color key
import math

# Get input_card and set global variables as "mcdc_"

import mcdc.global_ as mcdc_

input_card = mcdc_.input_deck


# get a point on the plane based on the current time in the system
# start and end times are zero by default
# called with a plane is visualized in create_cell_geometry()
def get_plane_current_position(surface, current_time, start_time, end_time):
    if len(surface["t"]) > 2:  # check if shape moves
        # establish reference points
        start_time = start_time  # default start time is zero
        end_time = end_time  # default end time is zero
        start_position = -surface["J"][0][0]
        end_position = -surface["J"][1][0]

        current_position = start_position + (
            ((end_position - start_position) / (end_time - start_time)) * current_time
        )

    else:
        current_position = -surface["J"][0][0]

    return current_position


# create the CSG geometry for a cell
# called by draw_geometry()
def create_cell_geometry(cell, current_time, surface_list, start_time, end_time):
    cell_shape_list = []  # list of shapes that make up the cell
    for i in range(0, len(cell["surface_IDs"])):
        surface_ID = cell["surface_IDs"][i]

        ## Check type of shape

        if surface_list[surface_ID]["type"] == "plane-x":
            # get reference point from the surface card
            point = (
                (
                    get_plane_current_position(
                        surface_list[surface_ID],
                        current_time,
                        start_time=start_time,
                        end_time=end_time,
                    )
                ),
                0,
                0,
            )

            # get normal vector from the surface card
            if cell["positive_flags"][i]:
                vector = (
                    -surface_list[surface_ID]["G"],
                    surface_list[surface_ID]["H"],
                    surface_list[surface_ID]["I"],
                )
            else:
                vector = (
                    surface_list[surface_ID]["G"],
                    surface_list[surface_ID]["H"],
                    surface_list[surface_ID]["I"],
                )

            # planes have to be intersected to achive the wanted visualization in ngsolve
            cell_shape_list.append([Plane(Pnt(point), Vec(vector)), "intersect"])

        elif surface_list[surface_ID]["type"] == "plane-y":
            # get reference point from the surface card
            point = (
                0,
                (
                    get_plane_current_position(
                        surface_list[surface_ID],
                        current_time,
                        start_time=start_time,
                        end_time=end_time,
                    )
                ),
                0,
            )

            # get normal vector from the surface card
            if cell["positive_flags"][i]:
                vector = (
                    surface_list[surface_ID]["G"],
                    -surface_list[surface_ID]["H"],
                    surface_list[surface_ID]["I"],
                )
            else:
                vector = (
                    surface_list[surface_ID]["G"],
                    surface_list[surface_ID]["H"],
                    surface_list[surface_ID]["I"],
                )

            # planes have to be intersected to achive the wanted visualization in ngsolve
            cell_shape_list.append(
                [Plane(Pnt(point), Vec(vector)).col([1, 0, 0]), "intersect"]
            )

        elif surface_list[surface_ID]["type"] == "plane-z":
            # get reference point from the surface card
            point = (
                0,
                0,
                (
                    get_plane_current_position(
                        surface_list[surface_ID],
                        current_time,
                        start_time=start_time,
                        end_time=end_time,
                    )
                ),
            )

            # get normal vector from the surface card
            if cell["positive_flags"][i]:
                vector = (
                    surface_list[surface_ID]["G"],
                    surface_list[surface_ID]["H"],
                    -surface_list[surface_ID]["I"],
                )
            else:
                vector = (
                    surface_list[surface_ID]["G"],
                    surface_list[surface_ID]["H"],
                    surface_list[surface_ID]["I"],
                )

            # planes have to be intersected to achive the wanted visualization in ngsolve
            cell_shape_list.append(
                [Plane(Pnt(point), Vec(vector)).col([1, 0, 0]), "intersect"]
            )

        elif surface_list[surface_ID]["type"] == "sphere":
            # Get the center point from the surface card
            x = surface_list[surface_ID]["G"] / -2
            y = surface_list[surface_ID]["H"] / -2
            z = surface_list[surface_ID]["I"] / -2
            point = (x, y, z)

            # get radius from the surface card
            radius = float(
                math.sqrt(
                    abs(surface_list[surface_ID]["J"][0, 0] - x**2 - y**2 - z**2)
                )
            )

            # Add or subtract the sphere based on the CSG input
            if cell["positive_flags"][i]:
                cell_shape_list.append([Sphere(Pnt(point), radius), "subtract"])

            else:
                cell_shape_list.append([Sphere(Pnt(point), radius), "add"])

        elif surface_list[surface_ID]["type"] == "cylinder-x":
            # the y and z points are used to define the central axis of the cylinder
            y = surface_list[surface_ID]["H"] / -2
            z = surface_list[surface_ID]["I"] / -2

            # radius of the circle formed by a cross section parallel to the yz-plane
            radius = float(
                math.sqrt(abs(surface_list[surface_ID]["J"][0, 0] - z**2 - y**2))
            )

            # Add or subtract the sphere based on the CSG input
            # in the NGSolve CSG cylinders are infinite along their central axis,
            # by default we make the central axis -100<=x<=100
            # if a larger central axis is needed change below
            if cell["positive_flags"][i]:
                cell_shape_list.append(
                    [Cylinder(Pnt(-100, y, z), Pnt(100, y, z), radius), "subtract"]
                )
            else:
                cell_shape_list.append(
                    [Cylinder(Pnt(-100, y, z), Pnt(100, y, z), radius), "add"]
                )
        elif surface_list[surface_ID]["type"] == "cylinder-y":
            # the x and z points are used to define the central axis of the cylinder
            x = surface_list[surface_ID]["G"] / -2
            z = surface_list[surface_ID]["I"] / -2

            # radius of the circle formed by a cross section parallel to the xz-plane
            radius = float(
                math.sqrt(abs(surface_list[surface_ID]["J"][0, 0] - z**2 - y**2))
            )

            # Add or subtract the sphere based on the CSG input
            # in the NGSolve CSG cylinders are infinite along their central axis,
            # by default we make the central axis -100<=y<=100
            # if a larger central axis is needed change below
            if cell["positive_flags"][i]:
                cell_shape_list.append(
                    [Cylinder(Pnt(x, -100, z), Pnt(x, 100, z), radius), "subtract"]
                )
            else:
                cell_shape_list.append(
                    [Cylinder(Pnt(x, -100, z), Pnt(x, 100, z), radius), "add"]
                )
        elif surface_list[surface_ID]["type"] == "cylinder-z":
            # the x and y points are used to define the central axis of the cylinder
            x = surface_list[surface_ID]["G"] / -2
            y = surface_list[surface_ID]["H"] / -2

            # radius of the circle formed by a cross section parallel to the xy-plane
            radius = float(
                math.sqrt(abs(surface_list[surface_ID]["J"][0, 0] - z**2 - y**2))
            )

            # Add or subtract the sphere based on the CSG input
            # in the NGSolve CSG cylinders are infinite along their central axis,
            # by default we make the central axis -100<=z<=100
            # if a larger central axis is needed change below
            if cell["positive_flags"][i]:
                cell_shape_list.append(
                    [Cylinder(Pnt(x, y, -100), Pnt(x, y, 100), radius), "subtract"]
                )
            else:
                cell_shape_list.append(
                    [Cylinder(Pnt(x, y, -100), Pnt(x, y, 100), radius), "add"]
                )

    # combine the shapes for each cell
    # for subtraction to work properly, it must be done in the end
    for i in range(0, len(cell_shape_list)):
        if not (cell_shape_list[i][1] == "subtract"):
            if "cell_geometry" not in locals():
                cell_geometry = cell_shape_list[i][0]
            elif cell_shape_list[i][1] == "intersect":
                cell_geometry = cell_geometry * cell_shape_list[i][0]

            elif cell_shape_list[i][1] == "add":
                cell_geometry = cell_geometry + cell_shape_list[i][0]

    for i in range(0, len(cell_shape_list)):
        if cell_shape_list[i][1] == "subtract":
            cell_geometry = cell_geometry - cell_shape_list[i][0]

    return cell_geometry  # return the finished CSG geometry


# visualizes the model at a specified time (current_time, type float)
# called by visualize()
def draw_Geometry(current_time, start_time, end_time, material_colors):
    # create lists that contain all cells and surfaces
    surface_list = input_card.surfaces
    cell_list = input_card.cells

    geo = CSGeometry()  # create the ngsolve geometry object

    # colors that should not be generated  by distinctipy(starts with visually unappleaning colors, manually set colors added later)
    # These colors are rgb values, more can be added by extending the list
    input_colors = [
        (1, 1, 1),  # white
        (0, 0, 0),  # black
    ]
    # list of materials that need colors to be generated (ie not water or the source)
    material_colors_to_generate = []

    # if the color of water and source are not set make them blue and green respectivly
    # add manually specified colors to input colors.
    for cell in cell_list:
        cell_material_name = cell["material_name"]
        if cell_material_name not in list(material_colors.keys()):
            if cell_material_name == "water":
                material_colors["water"] = [0, 0, 1]
                input_colors.append((0, 0, 1))
            elif cell_material_name == "source":
                material_colors["source"] = [0, 1, 0]
                input_colors.append((0, 1, 0))
            else:
                material_colors[cell_material_name] = None
                material_colors_to_generate.append(cell_material_name)

    # create n number of distinct colors where n
    # is the number of materials in material_colors_to_generate
    distinct_colors = distinctipy.get_colors(
        len(material_colors_to_generate), input_colors
    )
    for i in range(0, len(material_colors_to_generate)):
        material_colors[material_colors_to_generate[i]] = [
            distinct_colors[i][0],
            distinct_colors[i][1],
            distinct_colors[i][2],
        ]

    # cycle through the cells in the model
    for cell_index in range(0, len(cell_list)):
        cell = cell_list[cell_index]

        # create the geometry for the cell
        cell_geometry = create_cell_geometry(
            cell,
            current_time=current_time,
            surface_list=surface_list,
            start_time=start_time,
            end_time=end_time,
        )
        # add the cell geometry to the visualization
        geo.Add(
            cell_geometry.col(material_colors[cell["material_name"]]), transparent=True
        )

    # draw the visualization
    geo.Draw()
    Redraw()

    return material_colors


# displays the color key to the user
# called by visualize()
def create_color_key(root, color_key_dict):
    tk.Label(root, text="color key").grid(row=2, column=0)

    # for each material in the color_key_dict display
    # the material name and corresponding color to the user
    for material_index in range(0, len(color_key_dict)):
        # create label for the material name
        tk.Label(root, text=str(list(color_key_dict)[material_index]) + ":").grid(
            row=3 + material_index, sticky=tk.W
        )

        # canvas where color will be displayed
        canvas = tk.Canvas(root, width=200, height=len(color_key_dict) * 50)

        # switch from rgb to hex for tkinter
        rgb = list(color_key_dict.values())[material_index]
        rgb = [int(255 * rgb[0]), int(255 * rgb[1]), int(255 * rgb[2])]
        colorval = "#{0:02x}{1:02x}{2:02x}".format(rgb[0], rgb[1], rgb[2])

        # add rectangle to canvas with material color
        canvas.create_rectangle(10, 10, 70, 60, fill=colorval)

        # add canvas to the window
        canvas.grid(row=3 + material_index, column=1, sticky=tk.E)


# triggered when time slider or time spinbox changed
# it redraws the model at the new time
def time_slider_changed(current_time, start_time, end_time, material_colors):
    draw_Geometry(
        current_time=float(current_time),
        start_time=start_time,
        end_time=end_time,
        material_colors=material_colors,
    )


# creates the time slider
# called by visualize()
def create_time_slider(root, start_time, end_time, tick_interval, material_colors):
    root.title("Time Slider")

    time_label = tk.Label(root, text="Time")
    time_label.grid(row=0, column=0, columnspan=2)

    time_var = tk.StringVar(root, "0")

    time_scale = tk.Scale(
        root,
        from_=start_time,
        to=end_time,
        orient=tk.HORIZONTAL,
        resolution=tick_interval,
        variable=time_var,
        command=lambda event: time_slider_changed(
            event, start_time, end_time, material_colors=material_colors
        ),
        length=400,
    )
    time_scale.grid(row=1, column=1, columnspan=4)

    time_spinbox = tk.Spinbox(
        root,
        from_=start_time,
        to=end_time,
        textvariable=time_var,
        increment=tick_interval,
        command=lambda: time_slider_changed(
            time_var.get(), start_time, end_time, material_colors=material_colors
        ),
    )
    time_spinbox.grid(row=0, column=3)


# runs the visualization for a model
# start and end times are default zero
# called in input file
def visualize(start_time=0, end_time=0, tick_interval=1, material_colors={}):
    # The dependence for the visualizer are quite large and include
    # netgen (75MB), intel mkl (200MB), among others
    # These are configurable as optional dependencies
    try:
        # launches visualization window
        # must be inside this loop so it doesn't launch when the visualizer is imported
        import netgen.gui
    except ImportError as e:
        print("")
        print(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")
        print("MC/DC visualization error:")
        print("     Dependencies for visualization not installed")
        print("     To install optional dependencies needed for visualization:")
        print("         <pip install mcdc[viz]> ")
        print("")
        print(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")
        print("")

    color_key_dic = draw_Geometry(
        current_time=0,
        start_time=start_time,
        end_time=end_time,
        material_colors=material_colors,
    )

    # Set up tkinter window
    root = tk.Tk()
    if start_time != end_time:
        create_time_slider(root, start_time, end_time, tick_interval, color_key_dic)
    create_color_key(root, color_key_dic)
    root.mainloop()  # mainloop for tkinter window



import numpy as np
import matplotlib.pyplot as plt
import h5py


# =============================================================================
# Plot results
# =============================================================================

# Results
with h5py.File("output.h5", "r") as f:
    phi_avg = f["tally/flux/mean"][:]
    phi_sd = f["tally/flux/sdev"][:]
    k = f["k_cycle"][:]
    k_avg = f["k_mean"][()]
    k_sd = f["k_sdev"][()]
    rg = f["gyration_radius"][:]
    x = f["tally/grid/x"][:]
    y = f["tally/grid/y"][:]

dx = x[1] - x[0]
x_mid = 0.5 * (x[1:] + x[:-1])
y_mid = 0.5 * (y[1:] + y[:-1])
Y, X = np.meshgrid(x_mid, y_mid)

norm = np.sum(phi_avg)
phi_avg = phi_avg / norm / dx**2
phi_sd = phi_sd / norm / dx**2

phi_fast = np.zeros_like(phi_avg[0])
phi_thermal = np.zeros_like(phi_avg[0])
phi_fast_sd = np.zeros_like(phi_avg[0])
phi_thermal_sd = np.zeros_like(phi_avg[0])

for i in range(5):
    phi_fast += phi_avg[i]
    phi_fast_sd += np.square(phi_sd[i])
phi_fast_sd = np.sqrt(phi_fast_sd) / phi_fast
for i in range(5, 7):
    phi_thermal += phi_avg[i]
    phi_thermal_sd += np.square(phi_sd[i])
phi_thermal_sd = np.sqrt(phi_thermal_sd) / phi_thermal

print("k = %.5f +- %.5f" % (k_avg, k_sd))
# Plot
N_iter = len(k)
(p1,) = plt.plot(np.arange(1, N_iter + 1), k, "-b", label="MC")
(p2,) = plt.plot(
    np.arange(1, N_iter + 1), np.ones(N_iter) * k_avg, ":r", label="MC-avg"
)
plt.fill_between(
    np.arange(1, N_iter + 1),
    np.ones(N_iter) * (k_avg - k_sd),
    np.ones(N_iter) * (k_avg + k_sd),
    alpha=0.2,
    color="r",
)
plt.xlabel("Iteration #")
plt.ylabel(r"$k$")
plt.grid()
ax2 = plt.gca().twinx()
(p3,) = ax2.plot(np.arange(1, N_iter + 1), rg, "g--", label="GyRad")
plt.ylabel(r"Gyration radius [cm]")
lines = [p1, p2, p3]
plt.legend(lines, [l.get_label() for l in lines])
plt.show()

plt.pcolormesh(X, Y, phi_fast, shading="nearest")
plt.colorbar()
ax = plt.gca()
ax.set_aspect("equal")
plt.xlabel(r"$x$ [cm]")
plt.ylabel(r"$y$ [cm]")
plt.title(r"Fast neutron flux")
plt.show()

plt.pcolormesh(X, Y, phi_fast_sd, shading="nearest")
plt.colorbar()
ax = plt.gca()
ax.set_aspect("equal")
plt.xlabel(r"$x$ [cm]")
plt.ylabel(r"$y$ [cm]")
plt.title(r"Fast neutron flux stdev")
plt.show()

plt.pcolormesh(X, Y, phi_thermal, shading="nearest")
plt.colorbar()
ax = plt.gca()
ax.set_aspect("equal")
plt.xlabel(r"$x$ [cm]")
plt.ylabel(r"$y$ [cm]")
plt.title(r"Thermal neutron flux")
plt.show()

plt.pcolormesh(X, Y, phi_thermal_sd, shading="nearest")
plt.colorbar()
ax = plt.gca()
ax.set_aspect("equal")
plt.xlabel(r"$x$ [cm]")
plt.ylabel(r"$y$ [cm]")
plt.title(r"Thermal neutron flux stdev")
plt.show()
